import 'dart:math' as math;

import 'package:knt_bloc/knt_bloc.dart';
import 'package:knt_isolate/knt_isolate.dart';

import '../model/book.dart';
import '../repo/book_repo.dart';

class BookBloc extends BasePagingBloc<void, Book> with Frontend {
  final _books = <Book>[];
  bool _usingIsolator = false;

  BookBloc() {
    watch<PrepareIsolateEvent>(
      _handlePrepareIsolateEvent,
      transformer: droppable(),
    );
    watch<LoadBooksEvent>(_handleLoadBooks);
  }

  @override
  bool get needLog => true;

  @override
  Future<BaseResponsePaging<Book>> pagingInProgress(
      BaseRequestPaging<void> pagingQuery) async {
    await _loadBook();
    List<Book> partialBooks = _books.sublist(
      pagingQuery.offset,
      math.min(pagingQuery.offset + pagingQuery.limit, _books.length),
    );
    await Future.delayed(const Duration(milliseconds: 200));
    return BaseResponsePaging<Book>(dataList: partialBooks);
  }

  @override
  BasePagingResultState<Book> pagingStateInstance(
    List<Book> dataList,
    int offset, {
    String? tag,
  }) =>
      ShowBooksState(dataList, offset, tag: tag);

  Future<void> _handlePrepareIsolateEvent(
    PrepareIsolateEvent event,
    Emitter<BaseState> emit,
  ) async {
    await initBackend(initializer: _launch);
  }

  Future<void> _handleLoadBooks(
    LoadBooksEvent event,
    Emitter<BaseState> emit,
  ) async {
    _usingIsolator = event.usingIsolate;
    emit(const BaseLoadingState());
    await Future.delayed(const Duration(milliseconds: 200));
    await _loadBook(tag: event.tag);
  }

  static BookBlocBackend _launch(BackendArgument<void> argument) {
    return BookBlocBackend(argument);
  }

  Future<void> _loadBook({String? tag}) async {
    if (_usingIsolator) {
      final Maybe<List<Book>> maybe = await run(event: tag);
      if (maybe.hasValue) {
        _showBook(maybe.value);
      }
      if (maybe.hasError && maybe.error is KntException) {
        throw maybe.error;
      }
      return;
    }
    final loadedBook = await BookRepo().fetchBooks(simulateLongData: false);
    _showBook(loadedBook);
  }

  Future<void> _showBook(List<Book> books) async {
    _books.clear();
    _books.addAll(books);
  }

  @override
  void initActions() {}
}

class BookBlocBackend extends Backend {
  final _repo = BookRepo();

  BookBlocBackend(BackendArgument<void> argument) : super(argument: argument);

  Future<List<Book>> _loadBooks({required String event, void data}) =>
      _repo.fetchBooks(simulateException: true);

  @override
  void initActions() {
    whenEventCome((LoadBooksEvent).toString()).run(_loadBooks);
  }
}

// region EVENT
class PrepareIsolateEvent extends BaseEvent {}

class LoadBooksEvent extends BaseEvent {
  final bool usingIsolate;

  const LoadBooksEvent({
    this.usingIsolate = false,
  });
}
// endregion

// region STATE
class ShowBooksState extends BasePagingResultState<Book> {
  const ShowBooksState(
    List<Book> dataList,
    int currentOffset, {
    String? tag,
  }) : super(dataList, currentOffset, tag: tag);
}
// endregion
