import 'dart:async';

import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';

import '../bloc/blocs.dart';
import '../defined.dart';
import '../entity/exception.dart';

const _defaultRatioScrollOffsetToLoadMore = -0.06;
const _defaultLoadingItemCount = 3;

class PagingWidget<Q, D, B extends BasePagingBloc<Q, D>>
    extends StatefulWidget {
  const PagingWidget({
    Key? key,
    required this.itemBuilder,
    this.needQueryImmediately = true,
    this.queryData,
    this.limit = defaultLimitPaging,
    this.queryConditions,
    this.insertedWidgetPositions = const [],
    this.loadingItemBuilder,
    this.itemDividerBuilder,
    this.searchBarBuilder,
    this.insertedWidgetBuilder,
    this.emptyDataWidget,
    this.errorWidgetBuilder,
    this.scrollPhysics,
    this.loadingItemCount = _defaultLoadingItemCount,
    this.gridView = false,
    this.sliverGridDelegate,
    this.direction = Axis.vertical,
    this.padding,
    this.ratioScrollOffsetToLoadMore = _defaultRatioScrollOffsetToLoadMore,
    this.onPickedItem,
    this.needPaging = true,
    this.needPullToRefresh = true,
    this.needShrinkWrap = false,
  }) : super(key: key);

  /// Whether query data immediately
  final bool needQueryImmediately;

  /// Query data for each paging request, such as searched keywords
  final Q? queryData;

  /// Query limit
  final int limit;

  /// For filter/search condition before triggering query
  final bool Function(Q)? queryConditions;

  /// When we want insert other widgets into list in specific positions, using this
  final List<int> insertedWidgetPositions;

  /// Builder for items in list
  final Widget Function(BuildContext context, int index, D data) itemBuilder;

  /// Builder for divider between items
  final Widget Function(BuildContext context, int index)? itemDividerBuilder;

  /// Using this builder to build loading items in case [loadingItemCount] > 0
  final Widget Function(BuildContext context, int index)? loadingItemBuilder;

  /// Builder for search widget, if we need search/filter function
  final Widget Function(BuildContext context, Function(Q)? onChangedQuery)?
      searchBarBuilder;

  /// Builder for other widgets which we want to insert into list, in case [insertedWidgetPositions] is not empty
  final Widget Function(BuildContext context, int index)? insertedWidgetBuilder;

  /// Builder for error widget, which presents error result
  final Widget Function(BuildContext context, KntException exception)?
      errorWidgetBuilder;

  /// Widget presents empty result
  final Widget? emptyDataWidget;

  /// Scroll physic, default is [AlwaysScrollableScrollPhysics]
  final ScrollPhysics? scrollPhysics;

  /// Total loading items
  final int loadingItemCount;

  /// Should using GridView or not, set **true** if using
  final bool gridView;

  /// Only works if [gridView] = true
  final SliverGridDelegate? sliverGridDelegate;

  /// Scroll direction
  final Axis direction;

  /// Padding insets
  final EdgeInsets? padding;

  /// Bouncing offset to trigger loading more, calculate by
  /// this ratio * height (Axis.vertical) or ratio * width (Axis.horizontal).
  /// Negative value if we want to load data before users scroll to the end of content
  final double ratioScrollOffsetToLoadMore;

  /// Callback while users tap on item
  final Function(D item)? onPickedItem;

  /// Whether need querying paging or not
  final bool needPaging;

  /// Whether need pull to refresh or not
  final bool needPullToRefresh;

  /// Whether need shrink content to fit inside a column or not
  final bool needShrinkWrap;

  @override
  _PagingWidgetState<Q, D, B> createState() => _PagingWidgetState<Q, D, B>();
}

class _PagingWidgetState<Q, D, B extends BasePagingBloc<Q, D>>
    extends State<PagingWidget<Q, D, B>> {
  // Normal query function
  late List<D> _dataList;
  late Q? _queryData;
  late bool _isLoading;
  late List<int> _insertedWidgetPositions;

  @override
  void initState() {
    super.initState();
    _dataList = [];
    _queryData = widget.queryData;
    _isLoading = false;
    _insertedWidgetPositions = List.from(widget.insertedWidgetPositions);
    // Must be sorted before starting calculating any data index
    _insertedWidgetPositions.sort();
    if (widget.needQueryImmediately) {
      Future.delayed(Duration.zero, _onFirstQuery);
    }
  }

  @override
  void didUpdateWidget(covariant PagingWidget<Q, D, B> oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (_queryData != widget.queryData) {
      _queryData = widget.queryData;
      _onFirstQuery();
    }
  }

  @override
  Widget build(BuildContext context) {
    final hasSearchBar = widget.searchBarBuilder != null;
    final screenWidth = MediaQuery.of(context).size.width;
    final screenHeight = MediaQuery.of(context).size.height;
    return SizedBox(
      width: screenWidth,
      child: NotificationListener<ScrollNotification>(
        onNotification: (scrollInfo) {
          if (!widget.needPaging) {
            // Don't care about paging
            return false;
          }
          final scrollOffsetToLoadMore = widget.direction == Axis.vertical
              ? screenHeight * widget.ratioScrollOffsetToLoadMore
              : screenWidth * widget.ratioScrollOffsetToLoadMore;
          if (!_isLoading &&
              scrollInfo is ScrollUpdateNotification &&
              (scrollInfo.metrics.pixels - scrollInfo.metrics.maxScrollExtent >
                  scrollOffsetToLoadMore)) {
            _onLoadMoreResult();
          }
          return false;
        },
        child: Stack(
          children: [
            BlocBuilder<B, BaseState>(
              builder: (context, state) {
                _isLoading = false;
                if (state is BaseErrorState) {
                  return widget.errorWidgetBuilder
                          ?.call(context, state.exception) ??
                      const SizedBox();
                }
                if (state is BaseEmptyResultState) {
                  return widget.emptyDataWidget ?? const SizedBox();
                }
                if (state is BaseLoadingState) {
                  _isLoading = true;
                }
                if (state is BasePagingResultState<D>) {
                  _dataList = state.data;
                }
                final loadingItemCount =
                    _isLoading ? widget.loadingItemCount : 0;
                final totalDataItemCount = _dataList.length;
                final totalInsertItemCount = _insertedWidgetPositions.length;
                final totalWillShownItemCount =
                    totalDataItemCount + totalInsertItemCount;
                final totalItemCount =
                    totalWillShownItemCount + loadingItemCount;
                final calculateRealItemIndex = (relativeIndex) {
                  if (_insertedWidgetPositions.isEmpty) {
                    return relativeIndex;
                  }
                  int nearestInsertedItemIndex = -1;
                  for (var i = 0; i < totalInsertItemCount; i++) {
                    if (relativeIndex >= _insertedWidgetPositions[i]) {
                      nearestInsertedItemIndex = i;
                    }
                  }
                  int numberOfInsertedItemBefore = nearestInsertedItemIndex + 1;
                  if (numberOfInsertedItemBefore <= 0) {
                    // Not found any inserted widget before this data widget
                    return relativeIndex;
                  }
                  // This real index could be larger than data list, so we must handle this case
                  return relativeIndex - numberOfInsertedItemBefore;
                };
                final scrollPhysics = widget.scrollPhysics ??
                    (widget.needShrinkWrap
                        ? const ClampingScrollPhysics()
                        : const AlwaysScrollableScrollPhysics());
                final itemBuilder = (context, index) {
                  if (index >= totalWillShownItemCount) {
                    return widget.loadingItemBuilder?.call(context, index) ??
                        const SizedBox();
                  }
                  if (_insertedWidgetPositions.contains(index)) {
                    return widget.insertedWidgetBuilder?.call(context, index) ??
                        const SizedBox();
                  }
                  final realIndex = calculateRealItemIndex(index);
                  if (realIndex >= totalDataItemCount) {
                    return const SizedBox();
                  }
                  final dataItem = _dataList[realIndex];
                  return InkWell(
                    onTap: () => widget.onPickedItem?.call(dataItem),
                    child:
                        widget.itemBuilder.call(context, realIndex, dataItem),
                  );
                };
                if (widget.gridView) {
                  final gridView = GridView.builder(
                    shrinkWrap: widget.needShrinkWrap,
                    physics: scrollPhysics,
                    scrollDirection: widget.direction,
                    padding: widget.padding ?? const EdgeInsets.all(0.0),
                    gridDelegate: widget.sliverGridDelegate ??
                        const SliverGridDelegateWithFixedCrossAxisCount(
                          crossAxisCount: 2,
                          mainAxisSpacing: 12.0,
                          crossAxisSpacing: 12.0,
                        ),
                    itemBuilder: itemBuilder,
                    itemCount: totalItemCount,
                  );
                  return widget.needPullToRefresh
                      ? RefreshIndicator(
                          onRefresh: _onRefresh,
                          child: gridView,
                        )
                      : gridView;
                }
                final listView = ListView.separated(
                  shrinkWrap: widget.needShrinkWrap,
                  physics: scrollPhysics,
                  scrollDirection: widget.direction,
                  padding: widget.padding ?? const EdgeInsets.all(0.0),
                  itemBuilder: itemBuilder,
                  separatorBuilder: (BuildContext context, int index) {
                    return widget.itemDividerBuilder?.call(context, index) ??
                        const Divider();
                  },
                  itemCount: totalItemCount,
                );
                return widget.needPullToRefresh
                    ? RefreshIndicator(
                        onRefresh: _onRefresh,
                        child: listView,
                      )
                    : listView;
              },
            ),
            if (hasSearchBar)
              widget.searchBarBuilder!.call(context, _onChangedSearchText),
          ],
        ),
      ),
    );
  }

  Future<void> _onRefresh() async {
    _onFirstQuery();
  }

  void _onFirstQuery() {
    if (!mounted) {
      return;
    }
    context.read<B>().add(BasePagingEvent(
          queryData: _queryData,
          limit: widget.limit,
          isFirstRequest: true,
        ));
  }

  void _onLoadMoreResult() {
    if (!mounted) {
      return;
    }
    context.read<B>().add(BasePagingEvent(
          queryData: _queryData,
          limit: widget.limit,
        ));
  }

  void _onChangedSearchText(Q newQueryData) {
    // Check conditions at first
    final conditions = widget.queryConditions;
    if (conditions != null && !conditions.call(newQueryData)) {
      return;
    }
    _queryData = newQueryData;
    // With new query, we also need to re-query as old data is reset
    _onFirstQuery();
  }
}
